# csv2pskc.py - script to convert a CSV file to PSKC
#
# Copyright (C) 2018-2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Script to convert a CSV file to PSKC."""

from __future__ import annotations

import argparse
import base64
import csv
import datetime
import sys
from binascii import a2b_hex
from typing import Any, Iterator, TextIO

import dateutil.parser

import pskc
from pskc.scripts.util import (
    OutputFile, VersionAction, get_key, get_password)


epilog = '''
supported columns:
  id, serial, secret, counter, time_offset, time_interval, interval,
  time_drift, issuer, manufacturer, response_length, algorithm
And any other properties of pskc.key.Key instances.

Report bugs to <python-pskc-users@lists.arthurdejong.org>.
'''.strip()

# set up command line parser
parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description='Convert a CSV file to PSKC.', epilog=epilog)
parser.add_argument(
    'input', nargs='?', metavar='FILE', help='the CSV file to read')
parser.add_argument(
    '-V', '--version', action=VersionAction)
parser.add_argument(
    '-o', '--output', metavar='FILE',
    help='write PSKC to file instead of stdout')
parser.add_argument(
    '-c', '--columns', metavar='COL|COL:LABEL,..',
    help='list of columns or label to column mapping to import')
parser.add_argument(
    '--skip-rows', metavar='N', type=int, default=1,
    help='the number of rows before rows with key information start')
parser.add_argument(
    '-x', '--set', metavar='COL=VALUE', action='append',
    type=lambda x: x.split('=', 1), dest='extra_columns',
    help='add an extra value that is added to all key containers')
parser.add_argument(
    '-p', '--password', '--passwd', metavar='PASS/FILE',
    help='password to use for encrypting the PSKC file)')
parser.add_argument(
    '-s', '--secret', metavar='KEY/FILE',
    help='hex encoded encryption key or a file containing the binary key')
encodings = {
    'hex': a2b_hex,
    'base32': base64.b32decode,
    'base64': base64.b64decode,
}
parser.add_argument(
    '-e', '--secret-encoding', choices=sorted(encodings.keys()),
    help='encoding used for reading key material',
    default='hex')


def from_column(key: str, value: str, secret_encoding: str) -> bytes | str | datetime.datetime:
    """Convert a key value read from a CSV file in a format for PSKC."""
    # decode encoded secret
    if key == 'secret':
        return encodings[secret_encoding](value)  # type: ignore[no-any-return,operator]
    # convert dates to timestamps
    if key.endswith('_date'):
        return dateutil.parser.parse(value)
    return value


def open_csvfile(inputfile: TextIO) -> Iterator[Any]:
    """Open the CSV file, trying to detect the dialect."""
    # Guess dialect if possible and open the CSV file
    dialect: str | type[csv.Dialect] = 'excel'
    try:
        # seek before read to skip sniffing on non-seekable files
        inputfile.seek(0)
        try:
            dialect = csv.Sniffer().sniff(inputfile.read(1024))
        except Exception:  # pragma: no cover (very hard to test in doctest)
            pass
        inputfile.seek(0)
    except IOError:  # pragma: no cover (very hard to test in doctest)
        pass
    return csv.reader(inputfile, dialect)


def main() -> None:
    """Convert a CSV file to PSKC."""
    # parse command-line arguments
    args = parser.parse_args()
    # open the CSV file
    csvfile = open_csvfile(open(args.input, 'r') if args.input else sys.stdin)
    # figure out the meaning of the columns
    columns = []
    if args.skip_rows > 0:
        columns = [x.lower().replace(' ', '_') for x in next(csvfile)]
        for _ in range(args.skip_rows - 1):
            next(csvfile)
    if args.columns:
        if ':' in args.columns:
            # --columns is a list of mappings
            mapping = dict(
                (label.lower().replace(' ', '_'), key.lower())
                for label, key in (
                    column.split(':')
                    for column in args.columns.split(',')))
            columns = [mapping.get(column, column) for column in columns]
        else:
            # --columns is a list of columns
            columns = [x.lower() for x in args.columns.split(',')]
    # store rows in PSKC structure
    pskcfile = pskc.PSKC()
    for row in csvfile:
        data = dict(args.extra_columns or [])
        for column, value in zip(columns, row):
            for key in column.split('+'):
                if value and key not in ('', '-'):
                    data[key] = from_column(key, value, args.secret_encoding)
        pskcfile.add_key(**data)
    # encrypt the file if needed
    if args.secret:
        pskcfile.encryption.setup_preshared_key(key=get_key(args.secret))
    elif args.password:
        pskcfile.encryption.setup_pbkdf2(get_password(args.password))
    # write output PSKC file
    with OutputFile(args.output) as output:
        pskcfile.write(output)



# key.py - module for handling keys from pskc files
# coding: utf-8
#
# Copyright (C) 2014-2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Module that handles keys stored in PSKC files."""

from __future__ import annotations


import array
import binascii
from typing import Any, TYPE_CHECKING, cast

from pskc.policy import Policy

if TYPE_CHECKING:  # pragma: no cover (only for mypy)
    from pskc import PSKC
    from pskc.device import Device


class EncryptedValue:
    """A container for an encrypted value."""

    def __init__(self, cipher_value: bytes, mac_value: bytes | None, algorithm: str | None) -> None:
        self.cipher_value = cipher_value
        self.mac_value = mac_value
        self.algorithm = algorithm

    @classmethod
    def create(cls, pskc: PSKC, value: bytes | bytearray | str) -> EncryptedValue:
        """Construct an encrypted value from a plaintext value."""
        # force conversion to bytestring
        if not isinstance(value, (bytes, bytearray)):
            value = value.encode()
        cipher_value = pskc.encryption.encrypt_value(cast(bytes, value))
        mac_value = None
        if pskc.mac.algorithm:
            mac_value = pskc.mac.generate_mac(cipher_value)
        assert pskc.encryption.algorithm
        return cls(cipher_value, mac_value, pskc.encryption.algorithm)

    def get_value(self, pskc: PSKC) -> bytes:
        """Provide the decrypted value."""
        from pskc.exceptions import DecryptionError
        plaintext = pskc.encryption.decrypt_value(self.cipher_value, self.algorithm)
        # allow MAC over plaintext or ciphertext
        # (RFC 6030 implies MAC over ciphertext but older draft used
        # MAC over plaintext)
        if self.mac_value and self.mac_value not in (
                pskc.mac.generate_mac(self.cipher_value),
                pskc.mac.generate_mac(plaintext)):
            raise DecryptionError('MAC value does not match')
        return plaintext


class EncryptedIntegerValue(EncryptedValue):
    """Class representing an encrypted integer value."""

    @classmethod
    def create(cls, pskc: PSKC, value: int) -> EncryptedValue:  # type: ignore[override]
        """Construct an encrypted value from a plaintext value."""
        str_value = '%x' % value
        n = len(str_value)
        bytes_value = binascii.unhexlify(str_value.zfill(n + (n & 1)))
        return super(EncryptedIntegerValue, cls).create(pskc, bytes_value)

    def get_value(self, pskc: PSKC) -> int:  # type: ignore[override]
        """Provide the decrypted integer value."""
        value = super(EncryptedIntegerValue, self).get_value(pskc)
        # try to handle value as ASCII representation
        if value.isdigit():
            return int(value)
        # fall back to do big-endian decoding
        result = 0
        for x in array.array('B', value):
            result = (result << 8) + x
        return result


class DataTypeProperty:
    """A data descriptor that delegates actions to DataType instances."""

    def __init__(self, name: str, doc: str) -> None:
        self.name = name
        self.__doc__ = doc

    def __get__(self, obj: Key, objtype: type | None) -> Any:
        value = getattr(obj, '_' + self.name, None)
        if isinstance(value, EncryptedValue):
            return value.get_value(obj.device.pskc)
        else:
            return value

    def __set__(self, obj: Key, val: Any) -> None:
        setattr(obj, '_' + self.name, val)


class DeviceProperty:
    """A data descriptor that delegates actions to the Device instance."""

    def __init__(self, name: str) -> None:
        self.name = name

    def __get__(self, obj: Key, objtype: type | None) -> Any:
        return getattr(obj.device, self.name)

    def __set__(self, obj: Key, val: Any) -> None:
        setattr(obj.device, self.name, val)


class Key:
    """Representation of a single key from a PSKC file.

    Instances of this class provide the following properties:

      id: unique key identifier (should be constant between interchanges)
      algorithm: identifier of the PSKC algorithm profile (URI)
      secret: the secret key itself (binary form, automatically decrypted)
      counter: event counter for event-based OTP
      time_offset: time offset for time-based OTP algorithms (in intervals)
      time_interval: time interval for time-based OTP in seconds
      time_drift: device clock drift (negative means device is slow)
      issuer: party that issued the key
      key_profile: reference to pre-shared key profile information
      key_reference: reference to an external key
      friendly_name: human-readable name for the secret key
      key_userid: user distinguished name associated with the key
      algorithm_suite: additional algorithm characteristics (e.g. used hash)
      challenge_encoding: format of the challenge for CR devices
      challenge_min_length: minimum accepted challenge length by device
      challenge_max_length: maximum size challenge accepted by the device
      challenge_check: whether the device will check an embedded check digit
      response_encoding: format of the response the device will generate
      response_length: the length of the response of the device
      response_check: whether the device appends a Luhn check digit
      policy: reference to policy information (see Policy class)

    This class also provides access to the manufacturer, serial, model,
    issue_no, device_binding, start_date, expiry_date, device_userid and
    crypto_module properties of the Device class.
    """

    def __init__(self, device: Device) -> None:

        self.device = device

        self.id: str | None = None
        self.algorithm: str | None = None

        self.issuer: str | None = None
        self.key_profile: str | None = None
        self.key_reference: str | None = None
        self.friendly_name: str | None = None
        self.key_userid: str | None = None

        self.algorithm_suite: str | None = None

        self.challenge_encoding: str | None = None
        self.challenge_min_length: int | None = None
        self.challenge_max_length: int | None = None
        self.challenge_check: bool | None = None

        self.response_encoding: str | None = None
        self.response_length: int | None = None
        self.response_check: bool | None = None

        self.policy = Policy(self)

    secret: bytes | None = DataTypeProperty(  # type: ignore[assignment]
        'secret', 'The secret key itself.')
    counter: int | None = DataTypeProperty(  # type: ignore[assignment]
        'counter', 'An event counter for event-based OTP.')
    time_offset: int | None = DataTypeProperty(  # type: ignore[assignment]
        'time_offset',
        'A time offset for time-based OTP (number of intervals).')
    time_interval: int | None = DataTypeProperty(  # type: ignore[assignment]
        'time_interval', 'A time interval in seconds.')
    time_drift: int | None = DataTypeProperty(  # type: ignore[assignment]
        'time_drift', 'Device clock drift value (number of time intervals).')

    manufacturer = DeviceProperty('manufacturer')
    serial = DeviceProperty('serial')
    model = DeviceProperty('model')
    issue_no = DeviceProperty('issue_no')
    device_binding = DeviceProperty('device_binding')
    start_date = DeviceProperty('start_date')
    expiry_date = DeviceProperty('expiry_date')
    device_userid = DeviceProperty('device_userid')
    crypto_module = DeviceProperty('crypto_module')

    def check(self) -> bool:
        """Check if all MACs in the message are valid.

        This will return `None` if there is no MAC to be checked. It will return
        True if all the MACs match. If any MAC fails a
        `DecryptionError` exception is raised.
        """
        if all(x is not False for x in (
                self.secret, self.counter, self.time_offset,
                self.time_interval, self.time_drift)):
            return True
        assert False  # pragma: no cover (only for mypy)  # noqa: B011

    @property
    def userid(self) -> str:
        """User identifier (either the key or device userid)."""
        return self.key_userid or self.device_userid



# -*- coding: utf-8 -*-
#
# python-pskc documentation build configuration file, created by
# sphinx-quickstart
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

"""python-pskc documentation build configuration."""

import pskc


# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.autodoc', 'sphinx.ext.doctest', 'sphinx.ext.coverage',
    'sphinx.ext.intersphinx',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
source_encoding = 'utf-8'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = u'python-pskc'
copyright = u'2014-2025 Arthur de Jong'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = pskc.__version__
# The full version, including alpha/beta/rc tags.
release = version

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_*', '.svn', '.git']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['pskc.']

# Automatically generate stub pages for autosummary entries.
autosummary_generate = True


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'default'

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
html_last_updated_fmt = '%Y-%m-%d'

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# Output file base name for HTML help builder.
htmlhelp_basename = 'python-pskcdoc'


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('pskc2csv', 'pskc2csv', u'Convert a PSKC file to CSV',
     [u'Arthur de Jong'], 1),
    ('pskc2pskc', 'pskc2pskc', u'Convert a PSKC file to standard format',
     [u'Arthur de Jong'], 1),
    ('csv2pskc', 'csv2pskc', u'Convert a CSV file to PSKC',
     [u'Arthur de Jong'], 1),
]

# If true, show URL addresses after external links.
man_show_urls = True



# signature.py - module for handling signed XML files
# coding: utf-8
#
# Copyright (C) 2017-2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Module for handling signed PSKC files.

This module defines a Signature class that handles the signature checking,
keys and certificates.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover (only for mypy)
    from lxml.etree import _Element, _ElementTree

    from pskc import PSKC


def sign_x509(
    xml: _Element,
    key: bytes,
    certificate: str,
    algorithm: str | None = None,
    digest_algorithm: str | None = None,
    canonicalization_method: str | None = None,
) -> _Element:
    """Sign PSKC data using X.509 certificate and private key.

    xml: an XML document
    key: the private key in binary format
    certificate: the X.509 certificate
    """
    from signxml import XMLSigner, XMLSignatureProcessor, methods  # type: ignore[attr-defined]
    algorithm = algorithm or 'rsa-sha256'
    digest_algorithm = digest_algorithm or 'sha256'
    canonicalization_method = (
        canonicalization_method or
        getattr(XMLSignatureProcessor, 'default_c14n_algorithm', None) or
        'http://www.w3.org/2006/12/xml-c14n11')
    return XMLSigner(
        method=methods.enveloped,
        signature_algorithm=algorithm.rsplit('#', 1)[-1].lower(),
        digest_algorithm=digest_algorithm.rsplit('#', 1)[-1].lower(),
        c14n_algorithm=canonicalization_method,
    ).sign(xml, key=key, cert=certificate)


def verify_x509(tree: _Element, certificate: str | None = None, ca_pem_file: str | None = None) -> _Element:
    """Verify signature in PSKC data against a trusted X.509 certificate.

    :param certificate: a PEM encoded certificate that is used for verification
    :param ca_pem_file: the name of a file that contains a CA certificate

    The signature can be verified in three ways:

    * The signature has an embedded certificate that is signed by a CA that is
      configured in the system CA store. In this case neither `certificate` or
      `ca_pem_file` need to be specified
    * The signature was made  and a certificate was transmitted out-of-band.
      In this case the `certificate` argument needs to be present.
    * The signature has a certificate that is signed by a specific CA who's
      certificate was transmitted out-of-band. In this case the `ca_pem_file`
      is used to point to a CA certificate file (but a certificate needs to be
      embedded inside the PSKC file).

    This function will raise an exception when the validation fails.

    After calling this function a verified version of the PSKC file will
    be present in the :attr:`signed_pskc` attribute.
    """
    from signxml import XMLVerifier  # type: ignore[attr-defined]
    return XMLVerifier().verify(  # type: ignore[union-attr,return-value]
        tree, x509_cert=certificate,
        ca_pem_file=ca_pem_file,
    ).signed_xml


class Signature:
    """Class for handling signature checking of the PSKC file.

    Instances of this class provide the following properties:

      is_signed: boolean to indicate whether a signature is present
      algorithm: identifier of the signing algorithm used
      canonicalization_method: identifier of the XML canonicalization used
      digest_algorithm: algorithm used for creating the hash
      issuer: issuer of the certificate
      serial: serial number of the certificate
      key: key that will be used when creating a signed PSKC file
      certificate: the certificate that is embedded in the signature
      signed_pskc: a PSKC instance with the signed information
    """

    def __init__(self, pskc: PSKC) -> None:
        self.pskc = pskc
        self._algorithm: str | None = None
        self.canonicalization_method: str | None = None
        self.digest_algorithm: str | None = None
        self.issuer: str | None = None
        self.serial: str | None = None
        self.key: bytes | None = None
        self.certificate: str | None = None

    @property
    def is_signed(self) -> bool:
        """Test whether the PSKC file contains a signature.

        This method does not check whether the signature is valid but only if
        one was present in the PSKC file.
        """
        return bool(
            self.algorithm or self.canonicalization_method or
            self.digest_algorithm or self.issuer or self.certificate)

    @property
    def algorithm(self) -> str | None:
        """Provide the signing algorithm used."""
        if self._algorithm:
            return self._algorithm
        return None

    @algorithm.setter
    def algorithm(self, value: str | None) -> None:
        from pskc.algorithms import normalise_algorithm
        self._algorithm = normalise_algorithm(value)

    @property
    def signed_pskc(self) -> PSKC:
        """Provide the signed PSKC information."""
        if not hasattr(self, '_signed_pskc'):
            self.verify()
        return self._signed_pskc

    def verify(self, certificate: str | None = None, ca_pem_file: str | None = None) -> bool:
        """Verify signature in PSKC data against a trusted X.509 certificate.

        The signature can be verified in three ways:

        * The signature has an embedded certificate that is signed by a CA that is
          configured in the system CA store. In this case neither `certificate` or
          `ca_pem_file` need to be specified
        * The signature was made  and a certificate was transmitted out-of-band.
          In this case the `certificate` argument needs to be present.
        * The signature has a certificate that is signed by a specific CA who's
          certificate was transmitted out-of-band. In this case the `ca_pem_file`
          is used to point to a CA certificate file (but a certificate needs to be
          embedded inside the PSKC file).

        This function will raise an exception when the validation fails. The `certificate`
        is expected to be passed as a PEM encoded string. The `ca_pem_file` should point to a CA
        certificate store (PEM encoded file).

        After calling this function a verified version of the PSKC file will
        be present in the :attr:`signed_pskc` attribute.
        """
        from pskc import PSKC
        from pskc.parser import PSKCParser
        self.tree: _Element | _ElementTree[_Element]
        signed_xml = verify_x509(self.tree, certificate, ca_pem_file)
        pskc = PSKC()
        PSKCParser.parse_document(pskc, signed_xml)
        self._signed_pskc = pskc
        return True

    def sign(self, key: bytes, certificate: str | None = None) -> None:
        """Add an XML signature to the file.

        Set up a key and optionally a certificate that will be used to create an
        embedded XML signature when writing the file.

        This is a utility function that is used to configure the properties
        needed to create a signed PSKC file.
        """
        self.key = key
        self.certificate = certificate

    def sign_xml(self, xml: _Element) -> _Element:
        """Sign an XML document with the configured key and certificate."""
        assert self.key
        assert self.certificate
        return sign_x509(
            xml, self.key, self.certificate, self.algorithm,
            self.digest_algorithm, self.canonicalization_method)



# serialiser.py - PSKC file parsing functions
# coding: utf-8
#
# Copyright (C) 2016-2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Module for serialising PSKC files to XML."""

from __future__ import annotations

import base64
from typing import Callable, IO, Sequence, TYPE_CHECKING

from pskc.key import EncryptedIntegerValue, EncryptedValue
from pskc.xml import find, mk_elem, move_namespaces, reformat, tostring

if TYPE_CHECKING:  # pragma: no cover (only for mypy)
    from lxml.etree import _Element

    from pskc import PSKC
    from pskc.device import Device
    from pskc.encryption import Encryption, KeyDerivation
    from pskc.key import Key
    from pskc.mac import MAC
    from pskc.policy import Policy
    from pskc.signature import Signature


def my_b64encode(value: bytes | str) -> str:
    """Wrap around b64encode to handle types correctly."""
    if not isinstance(value, bytes):
        value = value.encode()
    return base64.b64encode(value).decode()


class PSKCSerialiser:
    """Class for serialising a PSKC structure to PSKC 1.0 XML."""

    @classmethod
    def serialise_file(
        cls,
        pskc: PSKC,
        output: IO[str] | IO[bytes],
    ) -> None:
        """Write the PSKC structure to the specified output file."""
        xml = tostring(cls.serialise_document(pskc))
        try:
            output.write(xml)  # type: ignore[arg-type]
        except TypeError:
            # fall back to writing as string
            output.write(xml.decode('utf-8'))  # type: ignore[call-overload]

    @classmethod
    def serialise_document(cls, pskc: PSKC) -> _Element:
        """Convert the PSKC structure to an element tree structure."""
        container: _Element = mk_elem('pskc:KeyContainer', Version='1.0', Id=pskc.id)  # type: ignore[assignment]
        cls.serialise_encryption(pskc.encryption, container)
        cls.serialise_mac(pskc.mac, container)
        for device in pskc.devices:
            cls.serialise_key_package(device, container)
        return cls.serialise_signature(pskc.signature, container)

    @classmethod
    def serialise_encryption(cls, encryption: Encryption, container: _Element) -> None:
        """Provide an XML element tree for the encryption information."""
        if all(x is None
               for x in (encryption.id, encryption.key_name, encryption.key,
                         encryption.derivation.algorithm)):
            return
        encryption_key: _Element = mk_elem(  # type: ignore[assignment]
            container, 'pskc:EncryptionKey', Id=encryption.id, empty=True)
        if encryption.derivation.algorithm:
            cls.serialise_key_derivation(
                encryption.derivation, encryption_key, encryption.key_names)
        else:
            for name in encryption.key_names:
                mk_elem(encryption_key, 'ds:KeyName', name)

    @classmethod
    def serialise_key_derivation(
        cls,
        derivation: KeyDerivation,
        encryption_key: _Element,
        key_names: Sequence[str],
    ) -> None:
        """Provide an XML structure for the key derivation properties."""
        assert derivation.algorithm
        derived_key = mk_elem(encryption_key, 'xenc11:DerivedKey', empty=True)
        key_derivation = mk_elem(derived_key, 'xenc11:KeyDerivationMethod',
                                 Algorithm=derivation.algorithm)
        if derivation.algorithm.endswith('#pbkdf2'):
            pbkdf2 = mk_elem(key_derivation, 'xenc11:PBKDF2-params',
                             empty=True)
            if derivation.pbkdf2_salt:
                salt = mk_elem(pbkdf2, 'Salt', empty=True)
                mk_elem(salt, 'Specified',
                        base64.b64encode(derivation.pbkdf2_salt).decode())
            mk_elem(pbkdf2, 'IterationCount', derivation.pbkdf2_iterations)
            mk_elem(pbkdf2, 'KeyLength', derivation.pbkdf2_key_length)
            mk_elem(pbkdf2, 'PRF', derivation.pbkdf2_prf)
        # TODO: serialise ReferenceList/DataReference
        for name in key_names:
            mk_elem(derived_key, 'xenc11:MasterKeyName', name)

    @classmethod
    def serialise_mac(cls, mac: MAC, container: _Element) -> None:
        """Provide an XML structure for the encrypted MAC key."""
        key_value = getattr(mac, '_key', None) or mac.pskc.encryption.key
        if not mac.algorithm and not key_value:
            return
        mac_method = mk_elem(
            container, 'pskc:MACMethod', Algorithm=mac.algorithm, empty=True)
        if not key_value:
            return
        # encrypt the mac key if needed
        if not isinstance(key_value, EncryptedValue):
            key_value = EncryptedValue.create(mac.pskc, key_value)
        # construct encrypted MACKey
        algorithm = key_value.algorithm or mac.pskc.encryption.algorithm
        cipher_value = key_value.cipher_value
        assert cipher_value
        if mac.pskc.encryption.iv:
            cipher_value = mac.pskc.encryption.iv + cipher_value
        mac_key = mk_elem(mac_method, 'pskc:MACKey', empty=True)
        mk_elem(mac_key, 'xenc:EncryptionMethod', Algorithm=algorithm)
        cipher_data = mk_elem(mac_key, 'xenc:CipherData', empty=True)
        mk_elem(cipher_data, 'xenc:CipherValue',
                base64.b64encode(cipher_value).decode())

    @classmethod
    def serialise_key_package(cls, device: Device, container: _Element) -> None:
        """Provide an XML structure for key package."""
        key_package = mk_elem(container, 'pskc:KeyPackage', empty=True)
        assert key_package is not None
        if any(x is not None
               for x in (device.manufacturer, device.serial, device.model,
                         device.issue_no, device.device_binding,
                         device.start_date, device.expiry_date,
                         device.device_userid)):
            device_info = mk_elem(key_package, 'pskc:DeviceInfo', empty=True)
            mk_elem(device_info, 'pskc:Manufacturer', device.manufacturer)
            mk_elem(device_info, 'pskc:SerialNo', device.serial)
            mk_elem(device_info, 'pskc:Model', device.model)
            mk_elem(device_info, 'pskc:IssueNo', device.issue_no)
            mk_elem(device_info, 'pskc:DeviceBinding', device.device_binding)
            mk_elem(device_info, 'pskc:StartDate', device.start_date)
            mk_elem(device_info, 'pskc:ExpiryDate', device.expiry_date)
            mk_elem(device_info, 'pskc:UserId', device.device_userid)
        if device.crypto_module is not None:
            crypto_module = mk_elem(key_package, 'pskc:CryptoModuleInfo',
                                    empty=True)
            mk_elem(crypto_module, 'pskc:Id', device.crypto_module)
        for key in device.keys:
            cls.serialise_key(key, key_package)

    @classmethod
    def serialise_key(cls, key: Key, key_package: _Element) -> None:
        """Provide an XML structure for the key information."""
        key_elm: _Element = mk_elem(  # type: ignore[assignment]
            key_package, 'pskc:Key', empty=True, Id=key.id, Algorithm=key.algorithm)
        mk_elem(key_elm, 'pskc:Issuer', key.issuer)
        if any((key.algorithm_suite, key.challenge_encoding,
                key.response_encoding, key.response_length)):
            parameters = mk_elem(key_elm, 'pskc:AlgorithmParameters',
                                 empty=True)
            mk_elem(parameters, 'pskc:Suite', key.algorithm_suite)
            mk_elem(parameters, 'pskc:ChallengeFormat',
                    Encoding=key.challenge_encoding,
                    Min=key.challenge_min_length,
                    Max=key.challenge_max_length,
                    CheckDigits=key.challenge_check)
            mk_elem(parameters, 'pskc:ResponseFormat',
                    Encoding=key.response_encoding,
                    Length=key.response_length,
                    CheckDigits=key.response_check)
        mk_elem(key_elm, 'pskc:KeyProfileId', key.key_profile)
        mk_elem(key_elm, 'pskc:KeyReference', key.key_reference)
        mk_elem(key_elm, 'pskc:FriendlyName', key.friendly_name)
        cls.serialise_data(
            key, 'secret', key_elm, 'pskc:Secret')
        cls.serialise_data(
            key, 'counter', key_elm, 'pskc:Counter')
        cls.serialise_data(
            key, 'time_offset', key_elm, 'pskc:Time')
        cls.serialise_data(
            key, 'time_interval', key_elm, 'pskc:TimeInterval')
        cls.serialise_data(
            key, 'time_drift', key_elm, 'pskc:TimeDrift')
        mk_elem(key_elm, 'pskc:UserId', key.key_userid)
        cls.serialise_policy(key.policy, key_elm)

    @classmethod
    def serialise_data(cls, key: Key, field: str, key_elm: _Element, tag: str) -> None:
        """Provide an XML structure for the key material."""
        value: bytes | str | EncryptedValue | None = getattr(key, '_%s' % field, None)
        pskc = key.device.pskc
        # skip empty values
        if value in (None, ''):
            return
        assert value is not None
        # get the value2text and encryption storage
        if field == 'secret':
            value2text: Callable[[bytes | str], str] = my_b64encode
            encrypted_value_cls = EncryptedValue
        else:
            value2text = str
            encrypted_value_cls = EncryptedIntegerValue
        # find the data tag and create our tag under it
        data = find(key_elm, 'pskc:Data')
        if data is None:
            data = mk_elem(key_elm, 'pskc:Data', empty=True)
        element = mk_elem(data, tag, empty=True)
        # see if we should encrypt the value
        if field in pskc.encryption.fields and not isinstance(value, EncryptedValue):
            value = encrypted_value_cls.create(pskc, value)
        # write out value
        if not isinstance(value, EncryptedValue):
            # unencrypted value
            mk_elem(element, 'pskc:PlainValue', value2text(value))
        else:
            # encrypted value
            algorithm = value.algorithm or pskc.encryption.algorithm
            cipher_value = value.cipher_value
            if pskc.encryption.iv:
                cipher_value = pskc.encryption.iv + cipher_value
            encrypted_value = mk_elem(
                element, 'pskc:EncryptedValue', empty=True)
            mk_elem(encrypted_value, 'xenc:EncryptionMethod',
                    Algorithm=algorithm)
            cipher_data = mk_elem(
                encrypted_value, 'xenc:CipherData', empty=True)
            mk_elem(cipher_data, 'xenc:CipherValue',
                    base64.b64encode(cipher_value).decode())
            if value.mac_value:
                mk_elem(element, 'pskc:ValueMAC',
                        base64.b64encode(value.mac_value).decode())

    @classmethod
    def serialise_policy(cls, policy: Policy, key_elm: _Element) -> None:
        """Provide an XML structure with the key policy information."""
        # check if any policy attribute is set
        if not policy.key_usage and all(x is None for x in (
                policy.start_date, policy.expiry_date,
                policy.number_of_transactions, policy.pin_key_id,
                policy.pin_usage, policy.pin_max_failed_attempts,
                policy.pin_min_length, policy.pin_max_length,
                policy.pin_encoding)):
            return
        policy_elm = mk_elem(key_elm, 'pskc:Policy', empty=True)
        mk_elem(policy_elm, 'pskc:StartDate', policy.start_date)
        mk_elem(policy_elm, 'pskc:ExpiryDate', policy.expiry_date)
        mk_elem(policy_elm, 'pskc:PINPolicy',
                PINKeyId=policy.pin_key_id,
                PINUsageMode=policy.pin_usage,
                MaxFailedAttempts=policy.pin_max_failed_attempts,
                MinLength=policy.pin_min_length,
                MaxLength=policy.pin_max_length,
                PINEncoding=policy.pin_encoding)
        for usage in policy.key_usage:
            mk_elem(policy_elm, 'pskc:KeyUsage', usage)
        mk_elem(policy_elm, 'pskc:NumberOfTransactions',
                policy.number_of_transactions)

    @classmethod
    def serialise_signature(cls, signature: Signature, container: _Element) -> _Element:
        """Provide an XML structure for embedded XML signature."""
        if not signature.key:
            return container
        # move the namespace to the root element and reformat before signing
        mk_elem(container, 'ds:Signature', Id='placeholder')
        container = move_namespaces(container)
        reformat(container)
        # sign the document
        return signature.sign_xml(container)



# util.py - utility functions for command-line scripts
#
# Copyright (C) 2014-2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Utility functions for command-line scripts."""

from __future__ import annotations

import argparse
import os.path
import sys
from binascii import a2b_hex
from typing import Any, TextIO

import pskc


version_string = '''
%s (python-pskc) %s
Written by Arthur de Jong.

Copyright (C) 2014-2025 Arthur de Jong
This is free software; see the source for copying conditions.  There is NO
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
'''.lstrip()


class VersionAction(argparse.Action):
    """Define --version argparse action."""

    def __init__(
        self,
        option_strings: list[str],
        dest: str,
        help: str = 'output version information and exit',
    ) -> None:
        super(VersionAction, self).__init__(
            option_strings=option_strings,
            dest=argparse.SUPPRESS,
            default=argparse.SUPPRESS,
            nargs=0,
            help=help,
        )

    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: Any,
        option_string: str | None = None,
    ) -> None:
        """Output version information and exit."""
        sys.stdout.write(version_string % (parser.prog, pskc.__version__))
        parser.exit()


def get_key(argument: str) -> bytes:
    """Get the key from a file or a hex-encoded string."""
    if os.path.isfile(argument):
        with open(argument, 'rb') as keyfile:
            return keyfile.read()
    else:
        return a2b_hex(argument)


def get_password(argument: str) -> str:
    """Get the password from a file or as a string."""
    if os.path.isfile(argument):
        with open(argument, 'r') as passfile:
            return passfile.read().replace('\n', '')
    else:
        return argument


class OutputFile:
    """Wrapper around output file to also fall back to stdout."""

    def __init__(self, output: str) -> None:
        self.output = output

    def __enter__(self) -> TextIO:
        self.file = open(self.output, 'w') if self.output else sys.stdout
        return self.file

    def __exit__(self, *args: Any) -> None:
        if self.output:
            self.file.close()
        else:  # we are using stdout
            self.file.flush()



# exceptions.py - collection of pskc exceptions
# coding: utf-8
#
# Copyright (C) 2014-2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Collection of exceptions."""

from __future__ import annotations


class PSKCError(Exception):
    """General top-level exception.

    The base class for all exceptions that the module will raise. In some
    cases third-party code may raise additional exceptions.
    """

    pass


class ParseError(PSKCError):
    """Something went wrong with parsing the PSKC file.

    Raised when the PSKC file cannot be correctly read due to invalid XML or
    some required element or attribute is missing. This exception should only
    be raised when parsing the file (i.e. when the :class:`~pskc.PSKC` class is
    instantiated).
    """

    pass


class EncryptionError(PSKCError):
    """There was a problem encrypting the value.

    Raised when encrypting a value is not possible due to key length issues,
    missing or wrong length plain text, or other issues.
    """

    pass


class DecryptionError(PSKCError):
    """There was a problem decrypting the value.

    Raised when decrypting a value fails due to missing or incorrect key,
    unsupported decryption or MAC algorithm, failed message authentication
    check or other error.

    This exception is generally raised when accessing encrypted information
    (i.e. the :attr:`~pskc.key.Key.secret`, :attr:`~pskc.key.Key.counter`,
    :attr:`~pskc.key.Key.time_offset`, :attr:`~pskc.key.Key.time_interval` or
    :attr:`~pskc.key.Key.time_drift` attributes of the :class:`~pskc.key.Key`
    class).
    """

    pass


class KeyDerivationError(PSKCError):
    """There was a problem performing the key derivation.

    Raised when key derivation fails due to an unsupported algorithm or
    missing information in the PSKC file.
    """

    pass



# encryption.py - module for handling encrypted values
# coding: utf-8
#
# Copyright (C) 2014-2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Module that handles encrypted PSKC values.

This module defines an Encryption class that handles the encryption key,
algorithms and decryption.

The encryption key can be derived using the KeyDerivation class.
"""

from __future__ import annotations

import os
import re
from collections.abc import Sequence
from typing import TYPE_CHECKING, Type

if TYPE_CHECKING:  # pragma: no cover (only for mypy)
    from cryptography.hazmat.primitives.ciphers import BlockCipherAlgorithm

    from pskc import PSKC


def algorithm_key_lengths(algorithm: str | None) -> Sequence[int]:
    """Return the possible key lengths for the configured algorithm."""
    from pskc.exceptions import DecryptionError
    if algorithm is None:
        raise DecryptionError('No algorithm specified')
    elif algorithm.endswith('#aes128-cbc') or \
            algorithm.endswith('#aes192-cbc') or \
            algorithm.endswith('#aes256-cbc'):
        return [int(algorithm[-7:-4]) // 8]
    elif algorithm.endswith('#tripledes-cbc') or \
            algorithm.endswith('#kw-tripledes'):
        return [16, 24]
    elif algorithm.endswith('#kw-aes128') or \
            algorithm.endswith('#kw-aes192') or \
            algorithm.endswith('#kw-aes256'):
        return [int(algorithm[-3:]) // 8]
    elif (algorithm.endswith('#camellia128-cbc') or
            algorithm.endswith('#camellia192-cbc') or
            algorithm.endswith('#camellia256-cbc')):
        return [int(algorithm[-7:-4]) // 8]
    elif (algorithm.endswith('#kw-camellia128') or
            algorithm.endswith('#kw-camellia192') or
            algorithm.endswith('#kw-camellia256')):
        return [int(algorithm[-3:]) // 8]
    else:
        raise DecryptionError('Unsupported algorithm: %r' % algorithm)


def _decrypt_cbc(
    algorithm: Type[BlockCipherAlgorithm],
    key: bytes,
    ciphertext: bytes,
    iv: bytes | None = None,
) -> bytes:
    """Decrypt the ciphertext and return the plaintext value."""
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import padding
    from cryptography.hazmat.primitives.ciphers import Cipher, modes
    from pskc.exceptions import DecryptionError
    if not iv:
        iv = ciphertext[:algorithm.block_size // 8]  # type: ignore[operator]
        ciphertext = ciphertext[algorithm.block_size // 8:]  # type: ignore[operator]
    cipher = Cipher(
        algorithm(key), modes.CBC(iv), backend=default_backend())  # type: ignore[call-arg]
    decryptor = cipher.decryptor()
    unpadder = padding.PKCS7(algorithm.block_size).unpadder()  # type: ignore[arg-type]
    try:
        return unpadder.update(
            decryptor.update(ciphertext) +
            decryptor.finalize()) + unpadder.finalize()
    except ValueError:
        raise DecryptionError('Invalid padding')


def decrypt(algorithm: str | None, key: bytes | None, ciphertext: bytes, iv: bytes | None = None) -> bytes:
    """Decrypt the ciphertext and return the plaintext value."""
    from cryptography.hazmat.primitives.ciphers import algorithms
    from pskc.exceptions import DecryptionError
    if key is None:
        raise DecryptionError('No key available')
    if algorithm is None:
        raise DecryptionError('No algorithm specified')
    if len(key) not in algorithm_key_lengths(algorithm):
        raise DecryptionError('Invalid key length')
    if algorithm.endswith('#aes128-cbc') or \
            algorithm.endswith('#aes192-cbc') or \
            algorithm.endswith('#aes256-cbc'):
        return _decrypt_cbc(algorithms.AES, key, ciphertext, iv)
    elif algorithm.endswith('#tripledes-cbc'):
        return _decrypt_cbc(algorithms.TripleDES, key, ciphertext, iv)
    elif algorithm.endswith('#kw-aes128') or \
            algorithm.endswith('#kw-aes192') or \
            algorithm.endswith('#kw-aes256'):
        from pskc.crypto.aeskw import unwrap as easkw_unwrap
        return easkw_unwrap(ciphertext, key)
    elif algorithm.endswith('#kw-tripledes'):
        from pskc.crypto.tripledeskw import unwrap as tripledeskw_unwrap
        return tripledeskw_unwrap(ciphertext, key)
    elif (algorithm.endswith('#camellia128-cbc') or
            algorithm.endswith('#camellia192-cbc') or
            algorithm.endswith('#camellia256-cbc')):
        return _decrypt_cbc(algorithms.Camellia, key, ciphertext, iv)
    elif (algorithm.endswith('#kw-camellia128') or  # pragma: no branch
            algorithm.endswith('#kw-camellia192') or
            algorithm.endswith('#kw-camellia256')):
        from pskc.crypto.aeskw import unwrap as easkw_unwrap
        return easkw_unwrap(ciphertext, key, algorithm=algorithms.Camellia)
    # no fallthrough because algorithm_key_lengths() fails with unknown algo
    assert False  # pragma: no cover (only for mypy)  # noqa: B011


def _encrypt_cbc(algorithm: Type[BlockCipherAlgorithm], key: bytes, plaintext: bytes, iv: bytes | None = None) -> bytes:
    """Encrypt the provided value with the key using the algorithm."""
    from cryptography.hazmat.backends import default_backend
    from cryptography.hazmat.primitives import padding
    from cryptography.hazmat.primitives.ciphers import Cipher, modes
    iv = iv or os.urandom(algorithm.block_size // 8)  # type: ignore[operator]
    cipher = Cipher(
        algorithm(key), modes.CBC(iv), backend=default_backend())  # type: ignore[call-arg]
    encryptor = cipher.encryptor()
    padder = padding.PKCS7(algorithm.block_size).padder()  # type: ignore[arg-type]
    return (
        iv + encryptor.update(
            padder.update(plaintext) + padder.finalize()) +
        encryptor.finalize())


def encrypt(algorithm: str | None, key: bytes | None, plaintext: bytes, iv: bytes | None = None) -> bytes:
    """Encrypt the provided value with the key using the algorithm."""
    from cryptography.hazmat.primitives.ciphers import algorithms
    from pskc.exceptions import EncryptionError
    if key is None:
        raise EncryptionError('No key available')
    if algorithm is None:
        raise EncryptionError('No algorithm specified')
    if len(key) not in algorithm_key_lengths(algorithm):
        raise EncryptionError('Invalid key length')
    if algorithm.endswith('#aes128-cbc') or \
            algorithm.endswith('#aes192-cbc') or \
            algorithm.endswith('#aes256-cbc'):
        return _encrypt_cbc(algorithms.AES, key, plaintext, iv)
    elif algorithm.endswith('#tripledes-cbc'):
        return _encrypt_cbc(algorithms.TripleDES, key, plaintext, iv)
    elif algorithm.endswith('#kw-aes128') or \
            algorithm.endswith('#kw-aes192') or \
            algorithm.endswith('#kw-aes256'):
        from pskc.crypto.aeskw import wrap as aeskw_wrap
        return aeskw_wrap(plaintext, key)
    elif algorithm.endswith('#kw-tripledes'):
        from pskc.crypto.tripledeskw import wrap as tripledeskw_wrap
        return tripledeskw_wrap(plaintext, key)
    elif (algorithm.endswith('#camellia128-cbc') or
            algorithm.endswith('#camellia192-cbc') or
            algorithm.endswith('#camellia256-cbc')):
        return _encrypt_cbc(algorithms.Camellia, key, plaintext, iv)
    elif (algorithm.endswith('#kw-camellia128') or  # pragma: no branch
            algorithm.endswith('#kw-camellia192') or
            algorithm.endswith('#kw-camellia256')):
        from pskc.crypto.aeskw import wrap as aeskw_wrap
        return aeskw_wrap(plaintext, key, algorithm=algorithms.Camellia)
    # no fallthrough because algorithm_key_lengths() fails with unknown algo
    assert False  # pragma: no cover (only for mypy)  # noqa: B011


class KeyDerivation:
    """Handle key derivation.

    The algorithm property contains the key derivation algorithm to use. For
    PBDKF2 the following parameters are set:

      pbkdf2_salt: salt value
      pbkdf2_iterations: number of iterations to use
      pbkdf2_key_length: required key length in bytes
      pbkdf2_prf: name of pseudorandom function used
    """

    def __init__(self) -> None:
        self._algorithm: str | None = None
        # PBKDF2 properties
        self.pbkdf2_salt: bytes | None = None
        self.pbkdf2_iterations: int | None = None
        self.pbkdf2_key_length: int | None = None
        self._pbkdf2_prf: str | None = None

    @property
    def algorithm(self) -> str | None:
        """Provide the key derivation algorithm used."""
        if self._algorithm:
            return self._algorithm
        return None

    @algorithm.setter
    def algorithm(self, value: str | None) -> None:
        from pskc.algorithms import normalise_algorithm
        self._algorithm = normalise_algorithm(value)

    @property
    def pbkdf2_prf(self) -> str | None:
        """Provide the PBKDF2 pseudorandom function used."""
        if self._pbkdf2_prf:
            return self._pbkdf2_prf
        return None

    @pbkdf2_prf.setter
    def pbkdf2_prf(self, value: str | None) -> None:
        from pskc.algorithms import normalise_algorithm
        self._pbkdf2_prf = normalise_algorithm(value)

    def derive_pbkdf2(self, password: str | bytes | bytearray) -> bytes:
        """Derive an encryption key from the provided password."""
        from hashlib import pbkdf2_hmac
        from pskc.exceptions import KeyDerivationError
        prf = 'sha1'
        if self.pbkdf2_prf:
            match = re.search(
                r'^(.*#)?hmac-(?P<hash>[a-z0-9-]+)$', self.pbkdf2_prf)
            if match:
                prf = match.group('hash')
            else:
                raise KeyDerivationError(
                    'Unsupported PRF: %r' % self.pbkdf2_prf)
        if not all((password, self.pbkdf2_salt, self.pbkdf2_key_length,
                    self.pbkdf2_iterations)):
            raise KeyDerivationError('Incomplete PBKDF2 configuration')
        # force conversion to bytestring
        if not isinstance(password, type(b'')):
            password = password.encode()  # type: ignore[union-attr]
        try:
            return pbkdf2_hmac(
                prf, password, self.pbkdf2_salt, self.pbkdf2_iterations,  # type: ignore[arg-type]
                self.pbkdf2_key_length)
        except ValueError:
            raise KeyDerivationError(
                'Pseudorandom function unsupported: %r' % self.pbkdf2_prf)

    def derive(self, password: str | bytes | bytearray) -> bytes:
        """Derive a key from the password."""
        from pskc.exceptions import KeyDerivationError
        if self.algorithm is None:
            raise KeyDerivationError('No algorithm specified')
        if self.algorithm.endswith('#pbkdf2'):
            return self.derive_pbkdf2(password)
        else:
            raise KeyDerivationError(
                'Unsupported algorithm: %r' % self.algorithm)

    def setup_pbkdf2(
        self,
        password: str | bytes | bytearray,
        salt: bytes | None = None,
        salt_length: int = 16,
        key_length: int | None = None,
        iterations: int | None = None,
        prf: str | None = None,
    ) -> bytes:
        """Configure PBKDF2 key derivation properties."""
        self.algorithm = 'pbkdf2'
        if salt is None:
            salt = os.urandom(salt_length)
        self.pbkdf2_salt = salt
        if iterations:
            self.pbkdf2_iterations = iterations
        elif self.pbkdf2_iterations is None:
            self.pbkdf2_iterations = 100000
        if key_length:  # pragma: no branch (always specified)
            self.pbkdf2_key_length = key_length
        if prf:
            self.pbkdf2_prf = prf
        return self.derive_pbkdf2(password)


class Encryption:
    """Class for handling encryption keys that are used in the PSKC file.

    Encryption generally uses a symmetric key that is used to encrypt some
    of the information stored in PSKC files (typically the seed). This
    class provides the following values:

      id: identifier of the key
      algorithm: the encryption algorithm used
      key_names: list of names for the key
      key_name: (first) name of the key (usually there is only one)
      key: the key value itself (binary form)
      iv: optional initialization vector for CBC based encryption
      fields: a list of Key fields that will be encrypted on writing

    The key can either be assigned to the key property or derived using the
    `derive_key()` method.
    """

    def __init__(self, pskc: PSKC) -> None:
        self.pskc = pskc
        self.id: str | None = None
        self._algorithm: str | None = None
        self.key_names: list[str] = []
        self.key: bytes | None = None
        self.iv: bytes | None = None
        self.derivation = KeyDerivation()
        self.fields: list[str] = []

    @property
    def key_name(self) -> str | None:
        """Provide the name of the (first) key."""
        if self.key_names:
            return self.key_names[0]
        return None

    @key_name.setter
    def key_name(self, value: str | None) -> None:
        if value:
            self.key_names = [value]
        else:
            self.key_names = []

    @property
    def algorithm(self) -> str | None:
        """Provide the encryption algorithm used."""
        if self._algorithm:
            return self._algorithm
        return None

    @algorithm.setter
    def algorithm(self, value: str | None) -> None:
        from pskc.algorithms import normalise_algorithm
        self._algorithm = normalise_algorithm(value)

    @property
    def is_encrypted(self) -> bool:
        """Test whether the PSKC file requires a decryption key."""
        from pskc.exceptions import DecryptionError
        try:
            for key in self.pskc.keys:
                key.secret, key.counter, key.time_offset
                key.time_interval, key.time_drift
        except DecryptionError:
            return True
        return False

    def derive_key(self, password: str | bytes | bytearray) -> None:
        """Derive a key from the password.

        The supplied password, together with the information embedded in the PSKC
        file (generally algorithm, salt, etc.) is used to create a decryption key.

        This function may raise a :exc:`~pskc.exceptions.KeyDerivationError`
        exception if key derivation fails.
        """
        self.key = self.derivation.derive(password)

    def _setup_encryption(
        self,
        *,
        id: str | None = None,
        algorithm: str | None = None,
        key_name: str | None = None,
        key_names: list[str] | None = None,
        fields: list[str] | None = None,
    ) -> None:
        if id is not None:
            self.id = id
        if algorithm is not None:
            self.algorithm = algorithm
        if key_name is not None:
            self.key_name = key_name
        if key_names is not None:
            self.key_names = key_names
        if fields is not None:
            self.fields = fields
        # default encryption to AES128-CBC
        if not self.algorithm:
            self.algorithm = 'aes128-cbc'
        # default to encrypting the secret only
        if not self.fields:
            self.fields = ['secret']
        # if we're using a CBC mode of encryption, add a MAC
        if self.algorithm.endswith('-cbc'):
            self.pskc.mac.setup()

    def setup_preshared_key(
        self,
        *,
        key: bytes | None = None,
        id: str | None = None,
        algorithm: str | None = None,
        key_length: int | None = None,
        key_name: str | None = None,
        key_names: list[str] | None = None,
        fields: list[str] | None = None,
    ) -> None:
        """Configure pre-shared key encryption when writing the file.

        :param key: the encryption key to use
        :param id: encryption key identifier
        :param algorithm: encryption algorithm
        :param key_length: encryption key length in bytes
        :param key_name: a name for the key
        :param key_names: a number of names for the key
        :param fields: a list of fields to encrypt

        This is a utility function to easily set up encryption. Encryption can
        also be set up by manually by setting the correct
        :class:`~pskc.encryption.Encryption` properties.

        This method will generate a key if required and set the passed values.
        By default AES128-CBC encryption will be configured and unless a key is
        specified one of the correct length will be generated. If the algorithm
        does not provide integrity checks (e.g. CBC-mode algorithms) integrity
        checking in the PSKC file will be set up using
        :func:`~pskc.mac.MAC.setup()`.

        By default only the :attr:`~pskc.key.Key.secret` property will be
        encrypted when writing the file.
        """
        self._setup_encryption(
            id=id,
            algorithm=algorithm,
            key_name=key_name,
            key_names=key_names,
            fields=fields,
        )
        if not key:
            key_length = key_length or self.algorithm_key_lengths[-1]
            key = os.urandom(key_length)
        self.key = key

    def setup_pbkdf2(
        self,
        password: str | bytes | bytearray,
        *,
        id: str | None = None,
        algorithm: str | None = None,
        key_name: str | None = None,
        key_names: list[str] | None = None,
        key_length: int | None = None,
        fields: list[str] | None = None,
        salt: bytes | None = None,
        salt_length: int = 16,
        iterations: int | None = None,
        prf: str | None = None,
    ) -> None:
        """Configure password-based PSKC encryption when writing the file.

        :param password: the password to use (required)
        :param id: encryption key identifier
        :param algorithm: encryption algorithm
        :param key_length: encryption key length in bytes
        :param key_name: a name for the key
        :param key_names: a number of names for the key
        :param fields: a list of fields to encrypt
        :param salt: PBKDF2 salt
        :param salt_length: used when generating random salt
        :param iterations: number of PBKDF2 iterations
        :param prf: PBKDF2 pseudorandom function

        Defaults for the above parameters are similar to those for
        :func:`setup_preshared_key()` but the password parameter is required.

        By default 12000 iterations will be used and a random salt with the
        length of the to-be-generated encryption key will be used.
        """
        self._setup_encryption(
            id=id,
            algorithm=algorithm,
            key_name=key_name,
            key_names=key_names,
            fields=fields,
        )
        self.key = self.derivation.setup_pbkdf2(
            password,
            salt=salt,
            salt_length=salt_length,
            key_length=key_length or self.algorithm_key_lengths[-1],
            iterations=iterations,
            prf=prf,
        )

    @property
    def algorithm_key_lengths(self) -> Sequence[int]:
        """Provide the possible key lengths for the configured algorithm."""
        return algorithm_key_lengths(self.algorithm)

    def decrypt_value(self, cipher_value: bytes, algorithm: str | None = None) -> bytes:
        """Decrypt the cipher_value and return the plaintext value."""
        return decrypt(
            algorithm or self.algorithm, self.key, cipher_value, self.iv)

    def encrypt_value(self, plaintext: bytes) -> bytes:
        """Encrypt the provided value and return the cipher_value."""
        cipher_value = encrypt(self.algorithm, self.key, plaintext, self.iv)
        if self.iv:
            cipher_value = cipher_value[len(self.iv):]
        return cipher_value

    def remove_encryption(self) -> None:
        """Decrypt all values and remove the encryption from the PSKC file.

        This can be used to read and encrypted PSKC file, decrypt the file,
        remove the encryption and output an unencrypted PSKC file or to replace
        the encryption algorithm.
        """
        # decrypt all values and store decrypted values
        for key in self.pskc.keys:
            key.secret = key.secret
            key.counter = key.counter
            key.time_offset = key.time_offset
            key.time_interval = key.time_interval
            key.time_drift = key.time_drift
        # remove MAC configuration
        self.pskc.mac.algorithm = None
        self.pskc.mac.key = None
        # remove encryption configuration
        self.id = None
        self.algorithm = None
        self.key_names = []
        self.key = None
        self.iv = None
        self.fields = []
        # remove key derivation configuration
        self.derivation.algorithm = None
        self.derivation.pbkdf2_salt = None
        self.derivation.pbkdf2_iterations = None
        self.derivation.pbkdf2_key_length = None
        self.derivation.pbkdf2_prf = None



# parser.py - PSKC file parsing functions
# coding: utf-8
#
# Copyright (C) 2016-2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Module for parsing PSKC files."""

from __future__ import annotations

import array
import base64
import copy
from typing import Any, IO, TYPE_CHECKING, cast

from pskc.encryption import Encryption, KeyDerivation
from pskc.exceptions import ParseError
from pskc.key import EncryptedIntegerValue, EncryptedValue
from pskc.mac import MAC
from pskc.xml import (
    find, findall, findbin, findint, findtext, findtime, getbool, getint,
    parse, remove_namespaces)

if TYPE_CHECKING:  # pragma: no cover (only for mypy)
    from os import PathLike

    from lxml.etree import _Element

    from pskc import PSKC
    from pskc.device import Device
    from pskc.key import Key
    from pskc.policy import Policy
    from pskc.signature import Signature


def plain2int(value: int | str | bytes) -> int:
    """Convert a plain text value to an int."""
    # try normal integer string parsing
    try:
        return int(value)
    except ValueError:
        pass
    # fall back to base64 decoding
    value = base64.b64decode(cast(bytes, value))
    # try to handle value as ASCII representation
    if value.isdigit():
        return int(value)
    # fall back to do big-endian decoding
    result = 0
    for x in array.array('B', value):
        result = (result << 8) + x
    return result


class PSKCParser:
    """Class to read various PSKC XML files into a PSKC structure."""

    @classmethod
    def parse_file(
        cls,
        pskc: PSKC,
        filename: str | bytes | PathLike[str] | PathLike[bytes] | IO[str] | IO[bytes],
    ) -> None:
        """Parse the provided file and store data in the PSKC instance."""
        try:
            tree = parse(filename)
        except Exception:
            raise ParseError('Error parsing XML')
        # save a clean copy of the tree for signature checking
        pskc.signature.tree = copy.deepcopy(tree)
        cls.parse_document(pskc, tree.getroot())

    @classmethod
    def parse_document(cls, pskc: PSKC, container: _Element) -> None:
        """Read information from the provided <KeyContainer> tree."""
        remove_namespaces(container)
        if container.tag not in ('KeyContainer', 'SecretContainer'):
            raise ParseError('Missing KeyContainer')
        # the version of the PSKC schema
        pskc.version = container.get('Version') or container.get('version')
        if (container.tag == 'KeyContainer' and
                pskc.version and
                pskc.version not in ('1', '1.0')):
            raise ParseError('Unsupported version %r' % pskc.version)
        # unique identifier for the container
        pskc.id = (
            container.get('Id') or container.get('ID') or container.get('id'))
        # handle EncryptionKey entries
        cls.parse_encryption(pskc.encryption, find(
            container, 'EncryptionKey', 'EncryptionMethod'))
        # handle MACMethod entries
        cls.parse_mac_method(pskc.mac, find(
            container, 'MACMethod', 'DigestMethod'))
        # fall back to MACAlgorithm
        mac_algorithm = findtext(container, 'MACAlgorithm')
        if mac_algorithm:
            pskc.mac.algorithm = mac_algorithm
        # handle KeyPackage entries
        for key_package in findall(container, 'KeyPackage', 'Device'):
            cls.parse_key_package(pskc.add_device(), key_package)
        # handle Signature entries
        cls.parse_signature(pskc.signature, find(container, 'Signature'))

    @classmethod
    def parse_encryption(cls, encryption: Encryption, key_info: _Element | None) -> None:
        """Read encryption information from the <EncryptionKey> XML tree."""
        if key_info is None:
            return
        encryption.id = key_info.get('Id')
        encryption.algorithm = (
            key_info.get('Algorithm') or
            key_info.get('algorithm') or
            encryption.algorithm)
        for name in findall(key_info,
                            'KeyName', 'DerivedKey/MasterKeyName',
                            'DerivedKey/CarriedKeyName'):
            encryption.key_names.append(findtext(name, '.'))  # type: ignore[arg-type]
        encryption.iv = findbin(key_info, 'IV') or encryption.iv
        cls.parse_key_derivation(encryption.derivation, find(
            key_info, 'DerivedKey/KeyDerivationMethod'))
        encryption.derivation.pbkdf2_salt = (
            findbin(key_info, 'PBESalt') or encryption.derivation.pbkdf2_salt)
        encryption.derivation.pbkdf2_iterations = (
            findint(key_info, 'PBEIterationCount') or
            encryption.derivation.pbkdf2_iterations)
        algorithm = (
            key_info.get('Algorithm') or key_info.get('algorithm') or '')
        if (algorithm.lower().startswith('pbe') and
                not encryption.derivation.algorithm):
            encryption.derivation.algorithm = 'pbkdf2'
            encryption.derivation.pbkdf2_key_length = (
                encryption.derivation.pbkdf2_key_length or
                encryption.algorithm_key_lengths[0])

    @classmethod
    def parse_key_derivation(cls, derivation: KeyDerivation, key_derivation: _Element | None) -> None:
        """Read derivation parameters from a <KeyDerivationMethod> element."""
        if key_derivation is None:
            return
        derivation.algorithm = key_derivation.get('Algorithm')
        # PBKDF2 properties
        pbkdf2 = find(key_derivation, 'PBKDF2-params')
        if pbkdf2 is not None:
            # get used salt
            derivation.pbkdf2_salt = findbin(pbkdf2, 'Salt/Specified')
            # required number of iterations
            derivation.pbkdf2_iterations = findint(pbkdf2, 'IterationCount')
            # key length
            derivation.pbkdf2_key_length = findint(pbkdf2, 'KeyLength')
            # pseudorandom function used
            prf = find(pbkdf2, 'PRF')
            if prf is not None:
                derivation.pbkdf2_prf = prf.get('Algorithm')

    @classmethod
    def parse_mac_method(cls, mac: MAC, mac_method: _Element | None) -> None:
        """Read MAC information from the <MACMethod> XML tree."""
        if mac_method is None:
            return
        mac.algorithm = (
            mac_method.get('Algorithm') or
            mac_method.get('algorithm'))
        mac_key = find(mac_method, 'MACKey')
        if mac_key is not None:
            algorithm, cipher_value = cls.parse_encrypted_value(mac_key)
            mac.key = EncryptedValue(cipher_value, None, algorithm)

    @classmethod
    def parse_key_package(cls, device: Device, key_package: _Element) -> None:
        """Read key information from the provided <KeyPackage> tree."""
        # find basic device information
        info = find(key_package, 'DeviceInfo', 'DeviceId')
        if info is not None:
            device.manufacturer = findtext(info, 'Manufacturer')
            device.serial = findtext(info, 'SerialNo')
            device.model = findtext(info, 'Model')
            device.issue_no = findtext(info, 'IssueNo')
            device.device_binding = findtext(info, 'DeviceBinding')
            device.start_date = findtime(info, 'StartDate')
            device.expiry_date = findtime(info, 'ExpiryDate', 'Expiry')
            device.device_userid = findtext(info, 'UserId')
        # find crypto module info
        device.crypto_module = findtext(key_package, 'CryptoModuleInfo/Id')
        # find keys for device
        for key_elm in findall(key_package, 'Key', 'Secret'):
            cls.parse_key(device.add_key(), key_elm)

    @classmethod
    def parse_key(cls, key: Key, key_elm: _Element) -> None:
        """Read key information from the provided <KeyPackage> tree."""
        # get key basic information
        key.id = (
            key_elm.get('Id') or key_elm.get('KeyId') or
            key_elm.get('SecretId'))
        key.algorithm = (
            key_elm.get('Algorithm') or key_elm.get('KeyAlgorithm') or
            key_elm.get('SecretAlgorithm'))
        # parse data section with possibly encrypted data
        data = find(key_elm, 'Data')
        if data is not None:
            cls.parse_data(key, 'secret', find(data, 'Secret'))
            cls.parse_data(key, 'counter', find(data, 'Counter'))
            cls.parse_data(key, 'time_offset', find(data, 'Time'))
            cls.parse_data(key, 'time_interval', find(data, 'TimeInterval'))
            cls.parse_data(key, 'time_drift', find(data, 'TimeDrift'))
        # parse legacy data elements with name attribute
        for data in findall(key_elm, 'Data'):
            name = data.get('Name')
            if name:
                cls.parse_data(key, dict(  # type: ignore[arg-type]
                    secret='secret',
                    counter='counter',
                    time='time_offset',
                    time_interval='time_interval',
                ).get(name.lower()), data)
        # parse more basic key properties
        key.issuer = findtext(key_elm, 'Issuer')
        key.key_profile = findtext(key_elm, 'KeyProfileId')
        key.key_reference = findtext(key_elm, 'KeyReference')
        key.friendly_name = findtext(key_elm, 'FriendlyName')
        # TODO: support multi-language values of <FriendlyName>
        key.key_userid = findtext(key_elm, 'UserId')
        key.algorithm_suite = findtext(
            key_elm, 'AlgorithmParameters/Suite')
        # parse challenge format
        challenge_format = find(
            key_elm,
            'AlgorithmParameters/ChallengeFormat', 'Usage/ChallengeFormat')
        if challenge_format is not None:
            key.challenge_encoding = (
                challenge_format.get('Encoding') or
                challenge_format.get('Format') or
                challenge_format.get('format'))
            key.challenge_min_length = (
                getint(challenge_format, 'Min') or
                getint(challenge_format, 'min'))
            key.challenge_max_length = (
                getint(challenge_format, 'Max') or
                getint(challenge_format, 'max'))
            key.challenge_check = getbool(
                challenge_format, 'CheckDigits', getbool(
                    challenge_format, 'CheckDigit'))
        # parse response format
        response_format = find(
            key_elm,
            'AlgorithmParameters/ResponseFormat', 'Usage/ResponseFormat')
        if response_format is not None:
            key.response_encoding = (
                response_format.get('Encoding') or
                response_format.get('Format') or
                response_format.get('format'))
            key.response_length = (
                getint(response_format, 'Length') or
                getint(response_format, 'length'))
            key.response_check = getbool(
                response_format, 'CheckDigits', getbool(
                    response_format, 'CheckDigit'))
        # parse key policy information
        cls.parse_policy(key.policy, find(key_elm, 'Policy'))
        # parse key usage information
        usage = find(key_elm, 'Usage')
        if usage is not None:
            for att in ('OTP', 'CR', 'Integrity', 'Encrypt', 'Unlock'):
                if getbool(usage, att):
                    key.policy.key_usage.append(att)
        key.policy.start_date = (
            findtime(key_elm, 'StartDate') or key.policy.start_date)
        key.policy.expiry_date = (
            findtime(key_elm, 'ExpiryDate') or key.policy.expiry_date)

    @classmethod
    def parse_encrypted_value(cls, encrypted_value: _Element) -> tuple[str | None, bytes]:
        """Read encryption value from <EncryptedValue> element."""
        algorithm = None
        cipher_value: bytes = findbin(encrypted_value, 'CipherData/CipherValue')  # type: ignore[assignment]
        encryption_method = find(encrypted_value, 'EncryptionMethod')
        if encryption_method is not None:
            algorithm = encryption_method.attrib.get('Algorithm')
        encryption_scheme = find(
            encrypted_value, 'EncryptionMethod/EncryptionScheme')
        if encryption_scheme is not None:
            algorithm = encryption_scheme.attrib.get('Algorithm') or algorithm
        return (algorithm, cipher_value)

    @classmethod
    def parse_data(cls, key: Key, field: str, element: _Element | None) -> None:
        """Read information from the provided element.

        The element is expected to contain <PlainValue>, <EncryptedValue>
        and/or <ValueMAC> elements that contain information on the actual
        value.
        """
        if element is None:
            return
        pskc = key.device.pskc
        plain_value: int | str | bytes | None = None
        cipher_value = None
        algorithm = None
        # get the plain2value function and encryption storage
        if field == 'secret':
            plain2value: Any = base64.b64decode
            encrypted_value_cls = EncryptedValue
        else:
            plain2value = plain2int
            encrypted_value_cls = EncryptedIntegerValue
        # read plaintext value from <PlainValue>
        plain_value = findtext(element, 'PlainValue')
        if plain_value is not None:
            plain_value = plain2value(plain_value)
        # read encrypted data from <EncryptedValue>
        encrypted_value = find(element, 'EncryptedValue')
        if encrypted_value is not None:
            algorithm, cipher_value = cls.parse_encrypted_value(
                encrypted_value)
            # store the found algorithm in the pskc.encryption property
            if not pskc.encryption.algorithm and algorithm:
                pskc.encryption.algorithm = algorithm
        # read MAC information from <ValueMAC>
        mac_value = findbin(element, 'ValueMAC', 'ValueDigest')
        # read legacy <Value> elements (can be plain or encrypted)
        value = findtext(element, 'Value')
        if value is not None:
            if pskc.encryption.algorithm and mac_value:
                cipher_value = findbin(element, 'Value')
            else:
                plain_value = plain2value(value)
        # store the found information
        if plain_value is not None:
            setattr(key, field, plain_value)
        elif cipher_value:
            setattr(key, field,
                    encrypted_value_cls(cipher_value, mac_value, cast(str, algorithm)))

    @classmethod
    def parse_policy(cls, policy: Policy, policy_elm: _Element | None) -> None:
        """Read key policy information from the provided <Policy> tree."""
        if policy_elm is None:
            return

        policy.start_date = findtime(policy_elm, 'StartDate')
        policy.expiry_date = findtime(policy_elm, 'ExpiryDate')
        policy.number_of_transactions = findint(
            policy_elm, 'NumberOfTransactions')
        for key_usage in findall(policy_elm, 'KeyUsage'):
            policy.key_usage.append(findtext(key_usage, '.'))  # type: ignore[arg-type]

        pin_policy_elm = find(policy_elm, 'PINPolicy')
        if pin_policy_elm is not None:
            policy.pin_key_id = pin_policy_elm.get('PINKeyId')
            policy.pin_usage = pin_policy_elm.get('PINUsageMode')
            policy.pin_max_failed_attempts = getint(
                pin_policy_elm, 'MaxFailedAttempts')
            policy.pin_min_length = getint(pin_policy_elm, 'MinLength')
            policy.pin_max_length = getint(pin_policy_elm, 'MaxLength')
            policy.pin_encoding = pin_policy_elm.get('PINEncoding')
            # check for child elements
            if list(pin_policy_elm):
                policy.unknown_policy_elements = True
            # check for unknown attributes
            known_attributes = set([
                'PINKeyId', 'PINUsageMode', 'MaxFailedAttempts', 'MinLength',
                'MaxLength', 'PINEncoding'])
            if set(pin_policy_elm.keys()) - known_attributes:
                policy.unknown_policy_elements = True

        # check for other child elements
        known_children = set([
            'StartDate', 'ExpiryDate', 'NumberOfTransactions', 'KeyUsage',
            'PINPolicy'])
        for child in policy_elm:
            if child.tag not in known_children:
                policy.unknown_policy_elements = True

    @classmethod
    def parse_signature(cls, signature: Signature, signature_elm: _Element | None) -> None:
        """Read signature information from the <Signature> element."""
        if signature_elm is None:
            return
        cm_elm = find(signature_elm, 'SignedInfo/CanonicalizationMethod')
        if cm_elm is not None:
            signature.canonicalization_method = cm_elm.attrib.get('Algorithm')
        sm_elm = find(signature_elm, 'SignedInfo/SignatureMethod')
        if sm_elm is not None:
            signature.algorithm = sm_elm.attrib.get('Algorithm')
        dm_elm = find(signature_elm, 'SignedInfo/Reference/DigestMethod')
        if dm_elm is not None:
            signature.digest_algorithm = dm_elm.attrib.get('Algorithm')
        issuer = find(signature_elm, 'KeyInfo/X509Data/X509IssuerSerial')
        if issuer is not None:
            signature.issuer = findtext(issuer, 'X509IssuerName')
            signature.serial = findtext(issuer, 'X509SerialNumber')
        certificate = findbin(
            signature_elm, 'KeyInfo/X509Data/X509Certificate')
        if certificate:
            certificate_str = base64.b64encode(certificate).decode('ascii')
            signature.certificate = '\n'.join(
                ['-----BEGIN CERTIFICATE-----'] +
                [certificate_str[i:i + 64]
                 for i in range(0, len(certificate_str), 64)] +
                ['-----END CERTIFICATE-----'])



# device.py - module for handling device info from pskc files
# coding: utf-8
#
# Copyright (C) 2016-2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Module that handles device information stored in PSKC files."""

from __future__ import annotations

import datetime
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover (only for mypy)
    from pskc import PSKC
    from pskc.key import Key


def update_attributes(obj: Key | Device, **kwargs: Any) -> None:
    """Update object with provided properties."""
    for k, v in kwargs.items():
        parts = k.split('.') if '.' in k else k.split('__')
        o = obj
        for name in parts[:-1]:
            o = getattr(o, name)
        getattr(o, parts[-1])  # raise exception for non-existing properties
        setattr(o, parts[-1], v)


class Device:
    """Representation of a single key from a PSKC file.

    Instances of this class provide the following properties:

      manufacturer: name of the organisation that made the device
      serial: serial number of the device
      model: device model description
      issue_no: issue number per serial number
      device_binding: device (class) identifier for the key to be loaded upon
      start_date: key should not be used before this date
      expiry_date: key or device may expire after this date
      device_userid: user distinguished name associated with the device
      crypto_module: id of module to which keys are provisioned within device
    """

    def __init__(self, pskc: PSKC) -> None:

        self.pskc = pskc

        self.manufacturer: str | None = None
        self.serial: str | None = None
        self.model: str | None = None
        self.issue_no: str | int | None = None
        self.device_binding: str | None = None
        self.start_date: datetime.datetime | None = None
        self.expiry_date: datetime.datetime | None = None
        self.device_userid: str | None = None
        self.crypto_module: str | None = None

        self.keys: list[Key] = []

    def add_key(self, **kwargs: str | bytes | int | datetime.datetime | None) -> Key:
        """Create a new key instance for the device.

        The new key is initialised with properties from the provided keyword
        arguments if any.

        The arguments may be any valid attribute of :class:`~pskc.key.Key` or
        be of the special form `policy.attribute` or `policy__attribute` to set
        any attribute of :class:`~pskc.policy.Policy`.
        """
        from pskc.key import Key
        key = Key(self)
        self.keys.append(key)
        update_attributes(key, **kwargs)
        return key



# pskc2pskc.py - script to convert a PSKC file to another PSKC file
#
# Copyright (C) 2018-2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Script to convert a PSKC file to another PSKC file."""

from __future__ import annotations

import argparse

import pskc
from pskc.scripts.util import (
    OutputFile, VersionAction, get_key, get_password)


epilog = '''
supported columns:
  id, serial, secret, counter, time_offset, time_interval, interval,
  time_drift, issuer, manufacturer, response_length, algorithm
And any other properties of pskc.key.Key instances.

Report bugs to <python-pskc-users@lists.arthurdejong.org>.
'''.strip()

# set up command line parser
parser = argparse.ArgumentParser(
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description='Convert a PSKC file to another PSKC file.', epilog=epilog)
parser.add_argument(
    'input', metavar='FILE', help='the PSKC file to read')
parser.add_argument(
    '-V', '--version', action=VersionAction)
parser.add_argument(
    '-o', '--output', metavar='FILE',
    help='write PSKC to file instead of stdout')
parser.add_argument(
    '-p', '--password', '--passwd', metavar='PASS/FILE',
    help='password to use for decryption (or read from a file)')
parser.add_argument(
    '-s', '--secret', metavar='KEY/FILE',
    help='hex encoded encryption key or a file containing the binary key')
parser.add_argument(
    '--new-password', '--new-passwd', metavar='PASS/FILE',
    help='password to use for encryption (or read from a file)')
parser.add_argument(
    '--new-secret', metavar='KEY/FILE',
    help='hex encoded encryption key or a file containing the binary key')


def main() -> None:
    """Convert a PSKC file to another PSKC file."""
    # parse command-line arguments
    args = parser.parse_args()
    # open and parse input PSKC file
    pskcfile = pskc.PSKC(args.input)
    if args.secret:
        pskcfile.encryption.key = get_key(args.secret)
        pskcfile.encryption.remove_encryption()
    elif args.password:
        pskcfile.encryption.derive_key(get_password(args.password))
        pskcfile.encryption.remove_encryption()
    # encrypt the output file if needed
    if args.new_secret:
        assert not pskcfile.encryption.is_encrypted
        pskcfile.encryption.setup_preshared_key(key=get_key(args.new_secret))
    elif args.new_password:
        assert not pskcfile.encryption.is_encrypted
        pskcfile.encryption.setup_pbkdf2(get_password(args.new_password))
    # write output PSKC file
    with OutputFile(args.output) as output:
        pskcfile.write(output)



# mac.py - module for checking value signatures
# coding: utf-8
#
# Copyright (C) 2014-2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Module that provides message authentication for PSKC values.

This module provides a MAC class that is used to store information about
how the MAC should be calculated (including the MAC key) and a ValueMAC
class that provides (H)MAC checking for PSKC key data.

The MAC key is generated specifically for each PSKC file and encrypted
with the PSKC encryption key.
"""

from __future__ import annotations

import hashlib
import os
import re

from pskc import PSKC
from pskc.key import EncryptedValue


_hmac_url_re = re.compile(r'^(.*#)?hmac-(?P<hash>[a-z0-9-]+)$')


def _get_hash_obj(algorithm: str | None, *args: bytes) -> hashlib._Hash:
    """Return an instantiated hash object."""
    from pskc.algorithms import normalise_algorithm
    from pskc.exceptions import DecryptionError
    match = _hmac_url_re.search(normalise_algorithm(algorithm) or '')
    if match:
        try:
            return hashlib.new(match.group('hash'), *args)
        except ValueError:
            pass
    raise DecryptionError('Unsupported MAC algorithm: %r' % algorithm)


def mac(algorithm: str | None, key: bytes | None, value: bytes) -> bytes:
    """Generate the MAC value over the specified value."""
    import hmac
    assert key
    return hmac.new(
        key, value,
        lambda *args: _get_hash_obj(algorithm, *args)).digest()


def mac_key_length(algorithm: str | None) -> int:
    """Recommended minimal key length in bytes for the set algorithm."""
    # https://tools.ietf.org/html/rfc2104#section-3
    # an HMAC key should be at least as long as the hash output length
    from pskc.exceptions import DecryptionError
    try:
        return int(_get_hash_obj(algorithm).digest_size)
    except DecryptionError:
        return 16  # fallback value


class MAC:
    """Class describing the MAC algorithm to use and how to get the key.

    Instances of this class provide the following attributes:

      algorithm: the name of the HMAC to use (currently only HMAC_SHA1)
      key: the binary value of the MAC key if it can be decrypted
    """

    def __init__(self, pskc: PSKC) -> None:
        self.pskc = pskc
        self._algorithm: str | None = None

    @property
    def key(self) -> bytes | None:
        """Provide access to the MAC key binary value if available."""
        value: bytes | EncryptedValue | None = getattr(self, '_key', None)
        if isinstance(value, EncryptedValue):
            return value.get_value(self.pskc)
        elif value:
            return value
        else:
            # fall back to encryption key
            return self.pskc.encryption.key

    @key.setter
    def key(self, value: bytes | EncryptedValue | None) -> None:
        self._key = value

    @property
    def algorithm(self) -> str | None:
        """Provide the MAC algorithm used."""
        if self._algorithm:
            return self._algorithm
        return None

    @algorithm.setter
    def algorithm(self, value: str | None) -> None:
        from pskc.algorithms import normalise_algorithm
        self._algorithm = normalise_algorithm(value)

    @property
    def algorithm_key_length(self) -> int:
        """Recommended minimal key length in bytes for the set algorithm."""
        return mac_key_length(self.algorithm)

    def generate_mac(self, value: bytes) -> bytes:
        """Generate the MAC over the specified value."""
        return mac(self.algorithm, self.key, value)

    def setup(self, key: bytes | None = None, algorithm: str | None = None) -> None:
        """Configure an encrypted MAC key.

        None of the arguments are required. By default HMAC-SHA1 will be used
        as a MAC algorithm. If no key is configured a random key will be
        generated with the length of the output of the configured hash.

        This function will automatically be called when the configured
        encryption algorithm requires a message authentication code.
        """
        if key:
            self.key = key
        if algorithm:
            self.algorithm = algorithm
        # default to HMAC-SHA1
        if not self.algorithm:
            self.algorithm = 'hmac-sha1'
        # generate an HMAC key
        if not self.key:
            self.key = os.urandom(self.algorithm_key_length)



# algorithms.py - module for handling algorithm URIs
# coding: utf-8
#
# Copyright (C) 2016-2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Utility module that handles algorithm URIs."""

from __future__ import annotations


# canonical URIs of known algorithms
# Note that even if a URI is listed here it does not mean that
# the algorithm is supported in python-pskc.
_algorithms = dict((x.rsplit('#', 1)[-1], x) for x in [
    'http://www.w3.org/2001/04/xmlenc#tripledes-cbc',
    'http://www.w3.org/2001/04/xmlenc#kw-tripledes',
    'http://www.w3.org/2001/04/xmlenc#arcfour',
    'http://www.w3.org/2001/04/xmlenc#aes128-cbc',
    'http://www.w3.org/2001/04/xmlenc#aes192-cbc',
    'http://www.w3.org/2001/04/xmlenc#aes256-cbc',
    'http://www.w3.org/2001/04/xmlenc#kw-aes128',
    'http://www.w3.org/2001/04/xmlenc#kw-aes192',
    'http://www.w3.org/2001/04/xmlenc#kw-aes256',
    'http://www.w3.org/2009/xmlenc11#aes128-gcm',
    'http://www.w3.org/2009/xmlenc11#aes192-gcm',
    'http://www.w3.org/2009/xmlenc11#aes256-gcm',
    'http://www.w3.org/2009/xmlenc11#kw-aes-128-pad',
    'http://www.w3.org/2009/xmlenc11#kw-aes-192-pad',
    'http://www.w3.org/2009/xmlenc11#kw-aes-256-pad',
    'http://www.w3.org/2001/04/xmldsig-more#camellia128-cbc',
    'http://www.w3.org/2001/04/xmldsig-more#camellia192-cbc',
    'http://www.w3.org/2001/04/xmldsig-more#camellia256-cbc',
    'http://www.w3.org/2001/04/xmldsig-more#kw-camellia128',
    'http://www.w3.org/2001/04/xmldsig-more#kw-camellia192',
    'http://www.w3.org/2001/04/xmldsig-more#kw-camellia256',
    'http://www.w3.org/2007/05/xmldsig-more#seed128-cbc'
    'http://www.w3.org/2007/05/xmldsig-more#kw-seed128',
    'http://www.w3.org/2001/04/xmldsig-more#hmac-md5',
    'http://www.w3.org/2000/09/xmldsig#hmac-sha1',
    'http://www.w3.org/2001/04/xmldsig-more#hmac-sha224',
    'http://www.w3.org/2001/04/xmldsig-more#hmac-sha256',
    'http://www.w3.org/2001/04/xmldsig-more#hmac-sha384',
    'http://www.w3.org/2001/04/xmldsig-more#hmac-sha512',
    'http://www.w3.org/2001/04/xmldsig-more#hmac-ripemd160',
    'http://www.rsasecurity.com/rsalabs/pkcs/schemas/pkcs-5v2-0#pbkdf2',
    # a newer URI seems to be http://www.w3.org/2009/xmlenc11#pbkdf2
])


# translation table to change old encryption names to new names
_algorithm_aliases = {
    '3des-cbc': 'tripledes-cbc',
    '3des112-cbc': 'tripledes-cbc',
    '3des168-cbc': 'tripledes-cbc',
    'rc4': 'arcfour',
    'kw-aes128-pad': 'kw-aes-128-pad',
    'kw-aes192-pad': 'kw-aes-192-pad',
    'kw-aes256-pad': 'kw-aes-256-pad',
    'camellia128': 'camellia128-cbc',
    'camellia192': 'camellia192-cbc',
    'camellia256': 'camellia256-cbc',
    'hmac-sha-1': 'hmac-sha1',
    'hmac-sha-224': 'hmac-sha224',
    'hmac-sha-256': 'hmac-sha256',
    'hmac-sha-384': 'hmac-sha384',
    'hmac-sha-512': 'hmac-sha512',
    'hmac-ripemd-160': 'hmac-ripemd160',
    'kw-3des': 'kw-tripledes',
    'pbe-3des112-cbc': 'tripledes-cbc',
    'pbe-3des168-cbc': 'tripledes-cbc',
    'pbe-aes128-cbc': 'aes128-cbc',
    'pbe-aes192-cbc': 'aes192-cbc',
    'pbe-aes256-cbc': 'aes256-cbc',
    'rsa-1_5': 'rsa-1_5',
    'rsa-oaep-mgf1p': 'rsa-oaep-mgf1p',
}


def normalise_algorithm(algorithm: str | None) -> str | None:
    """Return the canonical URI for the provided algorithm."""
    if not algorithm or algorithm.lower() == 'none':
        return None
    algorithm = _algorithm_aliases.get(algorithm.lower(), algorithm)
    return _algorithms.get(algorithm.rsplit('#', 1)[-1].lower(), algorithm)



# tripledeskw.py - implementation of Triple DES key wrapping
# coding: utf-8
#
# Copyright (C) 2014-2025 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Implement Triple DES key wrapping as described in RFC 3217."""

from __future__ import annotations

import binascii
import hashlib
import os

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from pskc.exceptions import DecryptionError, EncryptionError


def _cms_hash(value: bytes) -> bytes:
    """Return the key hash algorithm described in RFC 3217 section 2."""
    return hashlib.sha1(value).digest()[:8]


RFC3217_IV = binascii.a2b_hex('4adda22c79e82105')


def wrap(plaintext: bytes, key: bytes, iv: bytes | None = None) -> bytes:
    """Wrap one key (typically a Triple DES key) with another Triple DES key.

    This uses the algorithm from RFC 3217 to encrypt the plaintext (the key
    to wrap) using the provided key. If the iv is None, it is randomly
    generated.
    """
    if 8 * len(plaintext) % algorithms.TripleDES.block_size != 0:
        raise EncryptionError('Plaintext length wrong')
    if iv is None:
        iv = os.urandom(8)
    backend = default_backend()
    cipher = Cipher(algorithms.TripleDES(key), modes.CBC(iv), backend)
    encryptor = cipher.encryptor()
    tmp = (
        iv + encryptor.update(plaintext + _cms_hash(plaintext)) +
        encryptor.finalize())
    cipher = Cipher(algorithms.TripleDES(key), modes.CBC(RFC3217_IV), backend)
    encryptor = cipher.encryptor()
    return encryptor.update(tmp[::-1]) + encryptor.finalize()


def unwrap(ciphertext: bytes, key: bytes) -> bytes:
    """Unwrap a key (typically Triple DES key ) with another Triple DES key.

    This uses the algorithm from RFC 3217 to decrypt the ciphertext (the
    previously wrapped key) using the provided key.
    """
    if 8 * len(ciphertext) % algorithms.TripleDES.block_size != 0:
        raise DecryptionError('Ciphertext length wrong')
    backend = default_backend()
    cipher = Cipher(algorithms.TripleDES(key), modes.CBC(RFC3217_IV), backend)
    decryptor = cipher.decryptor()
    tmp = (decryptor.update(ciphertext) + decryptor.finalize())[::-1]
    cipher = Cipher(algorithms.TripleDES(key), modes.CBC(tmp[:8]), backend)
    decryptor = cipher.decryptor()
    tmp = decryptor.update(tmp[8:]) + decryptor.finalize()
    if tmp[-8:] == _cms_hash(tmp[:-8]):
        return tmp[:-8]
    raise DecryptionError('CMS key checksum error')
